
C $Id: fplane.f,v 3.4 1999/04/15 16:15:41 kohl Exp $


C
C                       CUMULVS version 1.0:
C       Computational Steering and Interactive Visualization
C                    in Distributed Applications
C            Oak Ridge National Laboratory, Oak Ridge TN.
C       Authors:  James Arthur Kohl and Philip M. Papadopoulos
C                    (C) 1996 All Rights Reserved
C
C                               NOTICE
C
C  Permission to use, copy, modify, and distribute this software and
C  its documentation for any purpose and without fee is hereby granted
C  provided that the above copyright notice appear in all copies and
C  that both the copyright notice and this permission notice appear
C  in supporting documentation.
C
C  Neither the Institution, Oak Ridge National Laboratory, nor the
C  Authors make any representations about the suitability of this
C  software for any purpose.  This software is provided ``as is''
C  without express or implied warranty.
C
C  CUMULVS was funded by the U.S. Department of Energy.
C


C ---	fplane.c - A better test program to test the packing
C ---	setup with CUMULVS.

		program fplane

		implicit none

		include 'fpvm3.h'

		include 'fstv.h'


C --- Parameter declarations

		character*7 FPLANEGRP

		integer CYCLESIZE
		integer IALLOC, JALLOC, KALLOC
		integer IDIM, JDIM, KDIM
		integer MAXPROCDIM
		integer NDIM
		integer OFFSETI, OFFSETJ, OFFSETK
		integer RINGMSG
		integer STVTAG

		logical VFLAG

		real*4 CENTERI, CENTERJ, CENTERK


C ---------  Adjustable Parameters  ----------


C --- Size of Global Computational Region is I_DIM X J_DIM X K_DIM

		parameter( NDIM = 3 )

		parameter( IDIM = 20 )
		parameter( JDIM = 30 )
		parameter( KDIM = 40 )


C --- Size of Array Allocation

		parameter( IALLOC = 100 )
		parameter( JALLOC = 100 )
		parameter( KALLOC = 100 )


C --- Misc Constants

		parameter( CYCLESIZE = 3 )

		parameter( FPLANEGRP = 'fplane' )

		parameter( MAXPROCDIM = 3 )

		parameter( RINGMSG = 1001 )

		parameter( STVTAG = 100 )


C --- Computational Parameters

		parameter( CENTERI = 0.0 )
		parameter( CENTERJ = 0.0 )
		parameter( CENTERK = 0.0 )

		parameter( OFFSETI = 10 )
		parameter( OFFSETJ = 20 )
		parameter( OFFSETK = 30 )

		parameter( vflag = .FALSE. )


C --- Main Program Variables

		real*4 density( IALLOC, JALLOC, KALLOC )

		real*4 BASE
		real*4 FACTOR
		real*4 INTERCEPT
		real*4 RISE
		real*4 SLOPE

		real*4 x, y, z
		real*4 value
		real*4 xx, zz
		real*4 sgn

		integer MUNG

		integer PSHAPE( MAXPROCDIM )
		integer PRANK

		integer NTASKS
		integer RESET

		integer glb( NDIM ), gub( NDIM )
		integer paddr( MAXPROCDIM )

		integer paramId( 10 )
		integer decompId
		integer fieldId

		integer globali, globalj, globalk
		integer locali, localj, localk
		integer dsttid, srctid
		integer begink, endk
		integer sequence
		integer i, j, k
		integer bstart
		integer cstart
		integer myinst
		integer block
		integer bufid
		integer cstep
		integer mytid
		integer slice
		integer idum
		integer info
		integer msrc
		integer qcnt
		integer q


C --- Processor Shape to Use

		data PSHAPE / 2, 3, 2 /
		data PRANK / 3 /


C --- Program Initialization

		call program_init( NTASKS, MAXPROCDIM, PSHAPE, PRANK,
     +		INTERCEPT, SLOPE, RISE, FACTOR, BASE, RESET )


C --- PVM Initialization

		call pvm_init( mytid, myinst, FPLANEGRP, RINGMSG, NTASKS,
     +		dsttid, srctid )


C --- Record the size of the computational region

		glb(1) = -IDIM / 2
		glb(2) = -JDIM / 2
		glb(3) = -KDIM / 2

		gub(1) = IDIM / 2
		gub(2) = JDIM / 2
		gub(3) = KDIM / 2


C --- Determine My Processor Address

		paddr(1) = myinst / ( PSHAPE(2) * PSHAPE(3) )
		paddr(2) = mod( myinst / PSHAPE(3), PSHAPE(2) )
		paddr(3) = mod( myinst, PSHAPE(3) )


C --- Determine Local Bounds for Explicit Axis

C --- Axis #1

		block = int( ( real( gub(1) - glb(1) + 1 )
     +		/ real( PSHAPE(1) ) ) + 1.0 )

		bstart = block * paddr(1)

		if ( vflag .eqv. .TRUE. ) then
			print *, 'block=', block, '  bstart=', bstart
		endif


C --- Axis #2

		cstart = CYCLESIZE * paddr(2)

		cstep = CYCLESIZE * PSHAPE(2)

		if ( vflag .eqv. .TRUE. ) then
			print *, 'cstart=', cstart, '  cstep=', cstep
		endif


C --- Axis #3

		slice = ( ( gub(3) - glb(3) + 1 ) - MUNG( PSHAPE(3) ) )
     +		/ PSHAPE(3)

		begink = glb(3) + ( paddr(3) * slice ) + MUNG( paddr(3) )

		endk = begink + slice + paddr(3)

		if ( endk .gt. gub(3) ) then
			endk = gub(3)
		endif

		if ( vflag .eqv. .TRUE. ) then
			print *, 'slice=', slice, '  begink=', begink,
     +			'  endk=', endk
		endif


C --- CUMULVS Initialization

		call cumulvs_init( myinst, FPLANEGRP, STVTAG, NTASKS,
     +		IALLOC, JALLOC, KALLOC, density, OFFSETI, OFFSETJ, OFFSETK,
     +		NDIM, glb, gub, MAXPROCDIM, paddr, PSHAPE, PRANK,
     +		SLOPE, INTERCEPT, RISE, BASE, FACTOR, RESET,
     +		CYCLESIZE, begink, endk, vflag,
     +		decompId, fieldId, paramId )


C --- Top of main computational loop */

		sequence = 1

10		continue

			if ( myinst .eq. 0 ) then
				print *, 'Sequence ', sequence
			endif

			do i = 0, block - 1

				globali = glb(1) + bstart + i

				locali = OFFSETI + i + 1

				q = cstart

				qcnt = 0

20				continue

					do j = 0, CYCLESIZE - 1

						globalj = glb(2) + q + j

						localj = OFFSETJ + ( qcnt * CYCLESIZE ) + j + 1

						if ( globalj .le. gub(2) ) then

							do k = 0, endk - begink

								globalk = begink + k

								localk = OFFSETK + k + 1

								x = real( globali ) - CENTERI
								y = real( globalj ) - CENTERJ
								z = real( globalk ) - CENTERK

								xx = ( x + y ) / 2.0

								zz = z - ( ( SLOPE * xx ) + INTERCEPT )

								if ( zz .ge. 0 ) then
									sgn = 1.0
								else
									sgn = -1.0
								endif

								value = RISE * sgn * real( sequence )
     +								* ( BASE ** ( FACTOR * abs( zz ) ) )

								if ( vflag .eqv. .TRUE. ) then

									print *, 'density[', locali,
     +									'][', localj, '][', localk,
     +									'] = f(', x, ',', y, ',', z,
     +									') = ', value

								endif

								density( locali, localj, localk ) =
     +								value

							enddo

						endif

					enddo

					q = q + cstep

					qcnt = qcnt + 1

				if ( q .lt. gub(2) - glb(2) + 1 ) goto 20

			enddo


C --- Send a message to mod( myinum + 1, ntasks )
C ---			Keep the tasks pseudo-synched

			call pvmfinitsend( PVMDATADEFAULT, info )

			call pvmfpack( INTEGER4, myinst, 1, 1, info )
			call pvmfpack( INTEGER4, mytid, 1, 1, info )

			call pvmfsend( dsttid, RINGMSG, info )

			call pvmfrecv( -1, RINGMSG, bufid )

			call pvmfbufinfo( bufid, idum, idum, msrc, info )

 			if ( msrc .ne. srctid ) then

				print *, 'Ring message seq ', sequence, ' from ',
     +				msrc, ' instead of ', srctid, ', exiting.'

				call pvmfbcast( FPLANEGRP, RINGMSG, info )

				call bail()

			endif

C --- Send messages to any attached CUMULVS front-end viewers

			call stvfsendtofe( info )

			if ( info .gt. 0 .and. myinst .eq. 0 ) then
				print *, 'stv sendToFE returned ', info
			endif

			call stvfsleep( 1 )

			if ( RESET .eq. 1 ) then
				sequence = 1
				RESET = 0
			else
				sequence = sequence + 1
			endif

		goto 10

		end


C --- Program Initialization Subroutine

		subroutine program_init( NTASKS, MAXPROCDIM, PSHAPE, PRANK,
     +		INTERCEPT, SLOPE, RISE, FACTOR, BASE, RESET )
		integer NTASKS
		integer MAXPROCDIM
		integer PSHAPE( MAXPROCDIM )
		integer PRANK
		real*4 INTERCEPT, SLOPE, RISE, FACTOR, BASE
		integer RESET

		integer i

		NTASKS = 1

		do i = 1, PRANK
			NTASKS = NTASKS * PSHAPE(i)
		enddo

		INTERCEPT = -3.0

		SLOPE = 1.2

		RISE = 2.0

		FACTOR = -0.75
		BASE = 1.1

		RESET = 0

		return

		end


C --- PVM Initialization Subroutine

		subroutine pvm_init( mytid, myinst, FPLANEGRP, RINGMSG, NTASKS,
     +		dsttid, srctid )

		include 'fpvm3.h'

		include 'fstv.h'

		integer mytid
		integer myinst
		character*7 FPLANEGRP
		integer RINGMSG
		integer NTASKS
		integer dsttid
		integer srctid

		integer tids(100)

		integer dst
		integer src

		integer nspawn
		integer info
		integer i


C --- Get My PVM tid and Join the FPLANE Group

		call pvmfmytid( mytid )

		call pvmfjoingroup( FPLANEGRP, myinst )


C --- Should I spawn other tasks?

		if ( myinst .eq. 0 ) then

			if ( NTASKS .gt. 1 ) then

				call pvmfcatchout( 1, info )

				call pvmfspawn( 'fplane',
     +				PVMTASKDEFAULT,
C    +				PVMTASKDEFAULT + PVMTASKDEBUG,
     +				' ', NTASKS - 1, tids, nspawn )

				if ( nspawn .lt. NTASKS - 1 ) then

					print *, 'Spawn failed with nspawn = ', nspawn

					print *, 'Killing orphaned tasks...'

					do i = 1, nspawn

						if ( tids(i) .lt. 0 ) then

							print *, 'Task #', i + 1, ': error = ',
     +							tids(i)

						else

							call pvmfkill( tids(i), info )
						
						endif

					enddo

					call bail()

				endif

			endif

		endif

C --- Make the group static -- this is also a barrier

		call pvmffreezegroup( FPLANEGRP, NTASKS, info )

C --- Determine src and destination for ring message

		dst = myinst + 1

		if ( dst .eq. NTASKS ) then
			dst = 0
		endif

		call pvmfgettid( FPLANEGRP, dst, dsttid )

		src = myinst - 1

		if ( src .lt. 0 ) then
			src = NTASKS - 1
		endif

		call pvmfgettid( FPLANEGRP, src, srctid )

C --- Notify me if my src task fails

		call pvmfnotify( PVMTASKEXIT, RINGMSG, 1, srctid, info )

		return

		end


C --- CUMULVS Initialization Subroutine

		subroutine cumulvs_init( myinst, FPLANEGRP, STVTAG, NTASKS,
     +		IALLOC, JALLOC, KALLOC, density, OFFSETI, OFFSETJ, OFFSETK,
     +		NDIM, glb, gub, MAXPROCDIM, paddr, PSHAPE, PRANK,
     +		SLOPE, INTERCEPT, RISE, BASE, FACTOR, RESET,
     +		CYCLESIZE, begink, endk, vflag,
     +		decompId, fieldId, paramId )

		include 'fpvm3.h'

		include 'fstv.h'

		integer myinst
		character*7 FPLANEGRP
		integer STVTAG
		integer NTASKS
		integer IALLOC, JALLOC, KALLOC
		real*4 density( IALLOC, JALLOC, KALLOC )
		integer OFFSETI, OFFSETJ, OFFSETK
		integer NDIM
		integer glb( NDIM )
		integer gub( NDIM )
		integer MAXPROCDIM
		integer paddr( MAXPROCDIM )
		integer PSHAPE( MAXPROCDIM )
		integer PRANK
		real*4 SLOPE, INTERCEPT, RISE, BASE, FACTOR
		integer RESET
		integer CYCLESIZE
		integer begink
		integer endk
		logical vflag
		integer decompId
		integer fieldId
		integer paramId( 10 )

		integer axisInfo1(3)
		integer axisInfo2(3)
		integer axisType(3)
		integer declare(3)
		integer offsets(3)
		integer info


C --- Initialize CUMULVS steering and visualization

		print *, 'Initializing CUMULVS...'
C		call stvfsetopt( STVOPTNULL, STVOPTVERBOSE, STVDEBUGALL, info )
		call stvfinit( FPLANEGRP, STVTAG, NTASKS, myinst, info )
		print *, '    info = ', info


C --- Set CUMULVS Options

		if ( vflag .eqv. .TRUE. ) then
			call stvfsetopt( STVOPTNULL, STVOPTVERBOSE,
     +			STVDEBUGALL, info )
		else
			call stvfsetopt( STVOPTNULL, STVOPTVERBOSE, 0, info )
		endif


C --- Define the steering parameters

		call stvfparamdefine( SLOPE, 'Slope', STVFLOAT,
     +		STVVISONLY, paramId(1) )

		call stvfparamdefine( INTERCEPT, 'Intercept', STVFLOAT,
     +		STVVISONLY, paramId(2) )

		call stvfparamdefine( RISE, 'Rise', STVFLOAT,
     +		STVVISONLY, paramId(3) )

		call stvfparamdefine( BASE, 'Exp.Base', STVFLOAT,
     +		STVVISONLY, paramId(4) )

		call stvfparamdefine( FACTOR, 'Exp.Factor', STVFLOAT,
     +		STVVISONLY, paramId(5) )

		call stvfparamdefine( RESET, 'Reset', STVINT,
     +		STVVISONLY, paramId(6) )


C --- Declare a decomposition of density
C ---		it is block, cyclic, explicit

		axisType(1) = STVBLOCK
		axisInfo1(1) = STVDEFAULT

		axisType(2) = STVCYCLIC
		axisInfo1(2) = CYCLESIZE

		axisType(3) = STVEXPLICIT
		axisInfo1(3) = begink
		axisInfo2(3) = endk

		call stvfdecompdefine( NDIM, axisType, axisInfo1, axisInfo2,
     +		glb, gub, PRANK, PSHAPE, decompId )


C --- Declare the density data field with the above decomposition

		declare(1) = IALLOC
		declare(2) = JALLOC
		declare(3) = KALLOC

		offsets(1) = OFFSETI
		offsets(2) = OFFSETJ
		offsets(3) = OFFSETK

		call stvffielddefine( density, 'density', decompId,
     +		offsets, declare, STVFLOAT, paddr, STVVISONLY,
     +		fieldId )

       call stvftimeof( 100, STVTIMEMILLISEC, tval )
       call stvfsetopt( fieldId, STVOPTFIELDTIMESCALE, tval, info )

C --- Verify Local Field Decomposition

		call stvfdumplocalfieldbounds( fieldId, STVVERBOSE, info )

		call stvfdumplocalfieldbounds( fieldId, STVTERSE, info )

		return

		end


C --- Bail-Out Subroutine

		subroutine bail()

		print *, 'Exiting PVM...'

		call pvmfexit( info )

		print *, 'Exiting'

		stop

		end


C --- Explicit Axis Munging Function

		integer function MUNG( n )
		integer n

		MUNG = n * ( n + 1 ) / 2

		return

		end

